/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  SERVER-STATUS
*
* @FILE     server-status.c
*
* @ABSTRACT
* Apache CCN Server Status  handler
*
* This file contains:
*
* @INCLUDE


/*
 * You can access the statistics with a URL like:
 *
 * http://your_server_name/server-status
 *
 * /server-status - Returns page using tables
 * /server-status?notable - Returns page for browsers without table support
 * /server-status?refresh - Returns page with 1 second refresh
 * /server-status?refresh=6 - Returns page with refresh every 6 seconds
 * /server-status?auto - Returns page with data for automatic parsing (XML)
*/

#define CORE_PRIVATE
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_conf_globals.h"	/* for ap_extended_status */
#include "http_main.h"
#include "util_script.h"
#include <time.h>
#include "scoreboard.h"
#include "http_log.h"
#if 0
#include "prelude.h"		/* iMatix sfl include file */
#include "sflxml.h"		/* iMatix sfl include file */
#include "sflxmll.h"		/* iMatix sfl include file */
#endif
#ifdef OK
#undef OK
#endif
#include "sfl.h"		/* iMatix sfl include file */
#include <server-status_upcon.h>

#define STATUS_MAXLINE		64

#define KBYTE			1024
#define	MBYTE			1048576L
#define	GBYTE			1073741824L

#ifndef DEFAULT_TIME_FORMAT 
#define DEFAULT_TIME_FORMAT "%A, %d-%b-%Y %H:%M:%S %Z"
#endif

module MODULE_VAR_EXPORT ccnServerStatus;

/*
 *command-related code. This is here to prevent use of ExtendedStatus
 * without ccnServerStatus included.
 */
static const char *set_extended_status(cmd_parms *cmd, void *dummy, int arg) 
{
    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
    if (err != NULL) {
        return err;
    }
    ap_extended_status = arg;
    return NULL;
}

static const command_rec status_module_cmds[] =
{
    { "ExtendedStatus", set_extended_status, NULL, RSRC_CONF, FLAG,
      "\"On\" to enable extended status information, \"Off\" to disable" },
    {NULL,NULL}
};

/* Format the number of bytes nicely */
static void format_byte_out(request_rec *r, unsigned long bytes)
{
    if (bytes < (5 * KBYTE))
	ap_rprintf(r, "%d B", (int) bytes);
    else if (bytes < (MBYTE / 2))
	ap_rprintf(r, "%.1f kB", (float) bytes / KBYTE);
    else if (bytes < (GBYTE / 2))
	ap_rprintf(r, "%.1f MB", (float) bytes / MBYTE);
    else
	ap_rprintf(r, "%.1f GB", (float) bytes / GBYTE);
}

static void format_kbyte_out(request_rec *r, unsigned long kbytes)
{
    if (kbytes < KBYTE)
	ap_rprintf(r, "%d kB", (int) kbytes);
    else if (kbytes < MBYTE)
	ap_rprintf(r, "%.1f MB", (float) kbytes / KBYTE);
    else
	ap_rprintf(r, "%.1f GB", (float) kbytes / MBYTE);
}

static void show_time(request_rec *r, time_t tsecs)
{
    long days, hrs, mins, secs;

    secs = tsecs % 60;
    tsecs /= 60;
    mins = tsecs % 60;
    tsecs /= 60;
    hrs = tsecs % 24;
    days = tsecs / 24;
    if (days)
	ap_rprintf(r, " %ld day%s", days, days == 1 ? "" : "s");
    if (hrs)
	ap_rprintf(r, " %ld hour%s", hrs, hrs == 1 ? "" : "s");
    if (mins)
	ap_rprintf(r, " %ld minute%s", mins, mins == 1 ? "" : "s");
    if (secs)
	ap_rprintf(r, " %ld second%s", secs, secs == 1 ? "" : "s");
}

/* Main handler for x-httpd-status requests */

/* ID values for command table */

#define STAT_OPT_END		-1
#define STAT_OPT_REFRESH	0
#define STAT_OPT_NOTABLE	1
#define STAT_OPT_AUTO		2

struct stat_opt {
    int id;
    const char *form_data_str;
    const char *hdr_out_str;
};

static const struct stat_opt status_options[] =	/* see #defines above */
{
    {STAT_OPT_REFRESH, "refresh", "Refresh"},
    {STAT_OPT_NOTABLE, "notable", NULL},
    {STAT_OPT_AUTO, "auto", NULL},
    {STAT_OPT_END, NULL, NULL}
};

static char status_flags[SERVER_NUM_STATUS];

static int StatusHandler(request_rec *r)
   {
   time_t nowtime = time(NULL);
   time_t up_time;
   int i, res;
   int ready = 0;
   int busy = 0;
   unsigned long count = 0;
   unsigned long lres, bytes;
   unsigned long bcount = 0;
   unsigned long kbcount = 0;
#ifndef NO_TIMES
#ifdef _SC_CLK_TCK
   float tick = sysconf(_SC_CLK_TCK);
#else
   float tick = HZ;
#endif
#endif
   int short_report = 1;
   int no_table_report = 0;
   short_score score_record;
   parent_score ps_record;
   char stat_buffer[HARD_SERVER_LIMIT];
   int pid_buffer[HARD_SERVER_LIMIT];
   unsigned long bytes_buffer[HARD_SERVER_LIMIT];
   unsigned long my_bytes_buffer[HARD_SERVER_LIMIT];
   clock_t tu, ts, tcu, tcs;
   char sztmp[20];
   char *pbuf;
   XML_ITEM *ccn_httpd, *httpd;

   tu = ts = tcu = tcs = 0;

   if (!ap_exists_scoreboard_image()) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
		    "Server status unavailable in inetd mode");
	return HTTP_INTERNAL_SERVER_ERROR;
   }
   r->allowed = (1 << M_GET);
   if (r->method_number != M_GET)
      return DECLINED;

   r->content_type = "text/html";

   ap_send_http_header(r);
#ifdef CHARSET_EBCDIC
   /* Server-generated response, converted */
   ap_bsetflag(r->connection->client, B_EBCDIC2ASCII, r->ebcdic.conv_out = 1);
#endif

   if (r->header_only)
      return 0;

   ap_sync_scoreboard_image();
   for (i = 0; i < HARD_SERVER_LIMIT; ++i) {
      score_record = ap_scoreboard_image->servers[i];
      ps_record = ap_scoreboard_image->parent[i];
      res = score_record.status;
      stat_buffer[i] = status_flags[res];
      pid_buffer[i] = (int) ps_record.pid;
       if (res == SERVER_READY)
         ready++;
      else if (res != SERVER_DEAD)
         busy++;
      lres = score_record.access_count;
      bytes_buffer[i] = bytes = score_record.bytes_served;
      my_bytes_buffer[i] = score_record.my_bytes_served;
      if (lres != 0 || (res != SERVER_READY && res != SERVER_DEAD)) {
#ifndef NO_TIMES
         tu += score_record.times.tms_utime;
         ts += score_record.times.tms_stime;
         tcu += score_record.times.tms_cutime;
         tcs += score_record.times.tms_cstime;
#endif /* NO_TIMES */
         count += lres;
         bcount += bytes;
         if (bcount >= KBYTE) {
            kbcount += (bcount >> 10);
            bcount = bcount & 0x3ff;
         }
      }
   }

   up_time = nowtime - ap_restart_time;

   ap_hard_timeout("send status info", r);

   /* Generate XML string */
   ccn_httpd = xml_new((XML_ITEM *)0, XML_CCN_STATUS_TAG, "");
   xml_put_attr_ic(ccn_httpd, XML_XMLNS_ATTR, XML_XMLNS_VALUE, FALSE);
   xml_put_attr_ic(ccn_httpd, XML_NAME_ATTR, XML_NAME_STATUS, FALSE);

   httpd = xml_new(ccn_httpd, XML_STATUS_TAG, ""); 
   sprintf(sztmp, "%lu", count);
   xml_put_attr_ic(httpd, XML_STATUS_TOTACCESS, sztmp, FALSE);
   sprintf(sztmp, "%lu", kbcount);
   xml_put_attr_ic(httpd, XML_STATUS_TOTKBYTES, sztmp, FALSE);
   if (ts || tu || tcu || tcs)
      {
      sprintf(sztmp, "%g", (tu + ts + tcu + tcs) / tick / up_time * 100.);
      xml_put_attr_ic(httpd, XML_STATUS_CPULOAD, sztmp, FALSE);
      }

   sprintf(sztmp, "%ld", (long) (up_time));
   xml_put_attr_ic(httpd, XML_STATUS_UPTIME, sztmp, FALSE);
   if (up_time > 0)
      {
      sprintf(sztmp, "%g", (float) count / (float) up_time);
      xml_put_attr_ic(httpd, XML_STATUS_REQPERSEC, sztmp, FALSE);
      sprintf(sztmp, "%g", KBYTE * (float) kbcount / (float) up_time);
      xml_put_attr_ic(httpd, XML_STATUS_BYTPERSEC, sztmp, FALSE);
      }

   if (count > 0)
      {
      sprintf(sztmp, "%g", KBYTE * (float) kbcount / (float) count);
      xml_put_attr_ic(httpd, XML_STATUS_BYTPERREC, sztmp, FALSE);
      }

   sprintf(sztmp, "%d", busy);
   xml_put_attr_ic(httpd, XML_STATUS_BUSYSERVERS, sztmp, FALSE);

   sprintf(sztmp, "%d", ready);
   xml_put_attr_ic(httpd, XML_STATUS_IDLESERVERS, sztmp, FALSE);

   httpd = xml_new(ccn_httpd, XML_CONFIG_TAG, ""); 
   sprintf(sztmp, "%d", ap_daemons_to_start);
   xml_put_attr_ic(httpd, XML_CONFIG_STARTSERVERS, sztmp, FALSE);

   sprintf(sztmp, "%d", r->server->keep_alive_max);
   xml_put_attr_ic(httpd, XML_CONFIG_MAXKPALIVREQ, sztmp, FALSE);

   sprintf(sztmp, "%d", ap_daemons_min_free);
   xml_put_attr_ic(httpd, XML_CONFIG_MINSPARSRV, sztmp, FALSE);

   sprintf(sztmp, "%d", ap_daemons_max_free);
   xml_put_attr_ic(httpd, XML_CONFIG_MAXSPARSRV, sztmp, FALSE);

   sprintf(sztmp, "%d", ap_daemons_limit);
   xml_put_attr_ic(httpd, XML_CONFIG_MAXCLIENT, sztmp, FALSE);

   sprintf(sztmp, "%d", ap_max_requests_per_child);
   xml_put_attr_ic(httpd, XML_CONFIG_MAXREQPCHILD, sztmp, FALSE);

   httpd = xml_new(ccn_httpd, XML_CHILDREN_TAG, ""); 
   for (i = 0; i < HARD_SERVER_LIMIT && pid_buffer[i] != 0; ++i)
      {
      XML_ITEM *httpd_child = xml_new(httpd, XML_CHILDOCC_TAG, ""); 
      sprintf(sztmp, "%c", stat_buffer[i]);
      xml_put_attr_ic(httpd_child, XML_CHILD_STATUS, sztmp, FALSE);

      sprintf(sztmp, "%d", pid_buffer[i]);
      xml_put_attr_ic(httpd_child, XML_CHILD_PID, sztmp, FALSE);

      sprintf(sztmp, "%lu", my_bytes_buffer[i] / MBYTE);
      xml_put_attr_ic(httpd_child, XML_CHILD_MYBYTES, sztmp, FALSE);

      sprintf(sztmp, "%lu", bytes_buffer[i] / MBYTE);
      xml_put_attr_ic(httpd_child, XML_CHILD_BYTES, sztmp, FALSE);
      }

   pbuf = xml_save_string(ccn_httpd);
   ap_rprintf(r, "%s%s\n", XML_HEADER, pbuf);
   mem_free(pbuf);
   xml_free(ccn_httpd);

#ifdef EAPI
   ap_hook_use("ap::status-handler::display",
                AP_HOOK_SIG4(void,ptr,int,int), AP_HOOK_ALL,
                r, no_table_report, short_report);
#endif


   ap_kill_timeout(r);
   return 0;
   }


static void status_init(server_rec *s, pool *p)
{
    status_flags[SERVER_DEAD] = '.';	/* We don't want to assume these are in */
    status_flags[SERVER_READY] = '_';	/* any particular order in scoreboard.h */
    status_flags[SERVER_STARTING] = 'S';
    status_flags[SERVER_BUSY_READ] = 'R';
    status_flags[SERVER_BUSY_WRITE] = 'W';
    status_flags[SERVER_BUSY_KEEPALIVE] = 'K';
    status_flags[SERVER_BUSY_LOG] = 'L';
    status_flags[SERVER_BUSY_DNS] = 'D';
    status_flags[SERVER_GRACEFUL] = 'G';
}

static const handler_rec status_handlers[] =
{
    {"status-handler", StatusHandler},
    {NULL,NULL}
};

module MODULE_VAR_EXPORT ccnServerStatus =
{
    STANDARD_MODULE_STUFF,
    status_init,		/* initializer */
    NULL,			/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    status_module_cmds,		/* command table */
    status_handlers,		/* handlers */
    NULL,			/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
#ifdef EAPI
    ,NULL
    ,NULL
    ,NULL
    ,NULL
    ,NULL
#endif
};

