/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
*
* @FILE     li_acmsc.c
*
* @ABSTRACT
* The Local Intelligence Access Control Management SCheduler.
* 
* This file contains:
*
* @INCLUDE
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gcn_upcon.h"
#include "gcn_uptyp.h"

#include "osx_upcon.h"
#include "osx_upext.h"

#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "gcn_upmes.h"
#include "li_prmes.h"

#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"

#include "gcn_upext.h"

#include "acm_upcon.h"
#include "acm_usext.h"

#include "rou_upext.h"

#include "li_prcon.h"
#include "li_prtyp.h"
#include "li_prext.h"
#include "li_upext.h"
#include "gcn_upmac.h"


/*
* @LOCALFUN li_ACLtoStr
*
* @DESCRIPTION Converts an Access Control List into a string
*
* @ABSTRACT
* This function converts an ACL (in fact, a CDXRESULT structure) into an
* argument string. The main utility of this routine is to convert the list
* of user profiles obtained by <@FUNCTION ACM_readACLuser>. Argument strings
* are handled by the GCN module, namely by the functions:
* <@FUNCTION gcn_AddArgToBuf> or <@FUNCTION gcn_GetNextBufArg>
*
* This function is called by <@LOCALFUN li_ACMScheduler>.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are : CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* IN     CDXRESULT *cdxACL_P
*        The Access Control List formatted as a CDXRESULT structure.
*
* OUT    CSICHAR **strACL_PP
*        The Access Control List formatted as an argument buffer.
*        This buffer is allocated by <@FUNCTION OS_mem_alloc>.
*
* OUT    CSILONG *returnCode_P
*        The returnCode.
*
* OUT    CSILONG *reasonCode_P
*        The reasonCode.
*
* @SEEALSO
* <@LOCALFUN li_ACMScheduler>
*/
static CSILONG li_ACLtoStr (
IN      CDXRESULT    *cdxACL_P,
OUT     CSICHAR     **strACL_PP,
OUT     CSILONG      *returnCode_P,
OUT     CSILONG      *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATION ***/
   /***********************************/

   CSILONG    ret       = CSISUCCESS;
   CSILONG    status    = LI_ERR_NONE;

   CSICHAR256 argWord   = {""};

   CSIULONG   strPos    = 0L;
   CSIULONG   strACLlen = 0L;

   CSICHAR   *tmpElem_P = NULL;

   TRACE_FUNCTION_ENTRY("li_ACLtoStr");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_ACLtoStr", CSIFAILED, LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** ARGUMENT BUFFER ALLOCATION ***/
   /**********************************/

   if (LI_ERR_NONE==status)
   {
      ret = OS_mem_alloc(GCN_MAX_CHAR_NUMBER+1,(CSIVOID **)strACL_PP,
                         returnCode_P, reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
         "OS_mem_alloc",(*reasonCode_P)));
         status = LI_ERR_ALLOC;
      }
      else
      {
         memset(*strACL_PP,0,(GCN_MAX_CHAR_NUMBER+1));
      }
   }

   /*** CONVERSION ***/
   /******************/

   if (LI_ERR_NONE==status)
   {
      tmpElem_P = CDX_list_first(NULL, cdxACL_P);

      while ( (NULL != tmpElem_P) && (LI_ERR_NONE == status) )
      {
         strncpy(argWord,tmpElem_P,sizeof(CSICHAR256));

         ret = gcn_AddArgToBuf(argWord,GCN_MAX_CHAR_NUMBER,
               GCN_SEPARATOR,strACL_PP,&strPos,&strACLlen,
               returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_AddArgToBuf",(*reasonCode_P)));
            status = LI_ERR_GCN_ARG;
         }
         tmpElem_P = CDX_list_next(NULL, cdxACL_P);
      }
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_ACLtoStr",*returnCode_P,*reasonCode_P);
   return(*returnCode_P);
}


/*
* @LOCALFUN li_ACMScheduler
*
* @DESCRIPTION Performs the Access Control (via the ACM module)
*
* @ABSTRACT
* The Local Intelligence Access Control Management Scheduler routine.
* This function relies on the Access Control Management library  to retrieve
* the User Profiles, the Access Control Lists and to check the matches.
*
* On the local gateway, the scheduler retrieves the user profiles (with
* <@FUNCTION ACM_readACLuser>).
*
* On the remote gateway, it retrieves the access control list of the target
* service or queue (with <@FUNCTION ACM_readACLservice> or
* <@FUNCTION ACM_readACLqueue>) and compares it with the user profiles
* (with <@FUNCTION ACM_checkACL>). It then accordingly grants or denies
* accesss to the concerned resources.
*
* This function is called by <@LOCALFUN li_OperationDispatcher>.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (failed).
*
* @PARAMETER
* IN     LI_PROCCONTEXT *li_ProcContext_P
*        The Local Intelligence process context.
*
* INOUT  LI_MSGCONTEXT *li_MsgContext_P
*        The context of the currently processed message.
*
* OUT    CSILONG *returnCode_P
*        The returnCode of the function, either CSISUCCESS or CSIFAILED.
*
* OUT    CSILONG *reasonCode_P
*        In case of error, precises the corresponding error code.
*
* @SEEALSO
* <@FUNCTION ACM_readACLuser>
* <@FUNCTION ACM_readACLservice>
* <@FUNCTION ACM_readACLqueue>
* <@FUNCTION ACM_checkACL>
* <@FUNCTION ACM_freeACL>
*/
CSILONG li_ACMScheduler (
IN      LI_PROCCONTEXT   *li_ProcContext_P,
INOUT   LI_MSGCONTEXT    *li_MsgContext_P,
OUT     CSILONG          *returnCode_P,
OUT     CSILONG          *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   CSILONG    ret         = CSISUCCESS;
   CSILONG    status      = LI_ERR_NONE;

   CSILONG    toLocal     = TRUE;

   CSICHAR   *tmp_P       = NULL;

   TRACE_FUNCTION_ENTRY("li_ACMScheduler");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_ACMScheduler",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** DESTINATION OF THE MESSAGE ***/
   /**********************************/

   /**********************************************************************/
   /* The following test enables us to know whether we are on the        */
   /* ORIGINATOR or TARGET gateway. Be aware that this might actually be */
   /* the same gateway, in case of a local to local transmission.        */
   /**********************************************************************/

   if (LI_ERR_NONE == status)
   {
      ret = CDX_isLocalGateway(li_MsgContext_P->DestGateway, &toLocal,
                               returnCode_P, reasonCode_P);
      if (CSIFAILED == ret)
      {
         status = *reasonCode_P;
      }
   }

   /*** RETRIEVAL OF USER PROFILES LIST ***/
   /***************************************/

   /***********************************************************************/
   /* We are on the ORIGINATOR gateway; we must retrieve the profiles of  */
   /* the user in the Directory; this is performed by ACM_readACLuser.    */
   /* The retrieved data is then converted to a string, so that it can be */
   /* transmitted in the trace information.                               */
   /***********************************************************************/

   if (LI_ERR_NONE == status)
   {
      if ( (0!=strcmp(li_MsgContext_P->TraceInfo_P->OrgUserName,""))&&
           (NULL==li_MsgContext_P->UserProfileList_P) )
      {  
         CDXRESULT  userACL = {CDXRESULT_DEFAULT};

         /* Retrieval of the User Profiles in the Directory */

         ret = ACM_readACLuser(li_MsgContext_P->TraceInfo_P->OrgUserName,
               &userACL, returnCode_P, reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_ACM_ERROR,
            "ACM_readACLuser",(*reasonCode_P)));
            status = LI_ERR_ACM;
         }
         else
         {
            /* Conversion CDXRESULT to String */
            /* The UserProfileList_P buffer is allocated by li_ACLtoStr() */
           
            ret = li_ACLtoStr(&userACL,&(li_MsgContext_P->UserProfileList_P),
                  returnCode_P, reasonCode_P);
            if (CSIFAILED==ret)
            {
               status = *reasonCode_P;
            }
            else
            {
               /* Counting the User Profiles */

               li_MsgContext_P->TraceInfo_P->NbOfOrgUserProfiles = 0L;

               tmp_P = CDX_list_first(NULL,&userACL);
               while (NULL != tmp_P)
               {
                 (li_MsgContext_P->TraceInfo_P->NbOfOrgUserProfiles)++;
                 tmp_P = CDX_list_next(NULL,&userACL);
               }
            }
         }
         CDX_free(NULL,&userACL);
      }
   }

   /*** IF TOLOCAL, PERFORM ACCESS CONTROL ***/
   /******************************************/

   /*******************************************************************/
   /* We are now on the TARGET gateway, and may proceed to the access */
   /* control. The user profiles are available as a string which was  */
   /* transmitted via the trace information. Access Control consists  */
   /* in 2 steps: first retrieving the ACL of the target service or   */
   /* queue, then comparing it with the User Profiles.                */
   /*******************************************************************/

   if (TRUE==toLocal)
   {
      CDXRESULT trgtACL = {CDXRESULT_DEFAULT};
      CDXRESULT tmpUsPr = {CDXRESULT_DEFAULT};

      if (LI_ERR_NONE == status)
      {
         if (CSI_SYNCHRON==li_MsgContext_P->ExchangeMode)
         {
            ret = ACM_readACLservice(
                  li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                  &trgtACL, returnCode_P, reasonCode_P);
            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_ACM_ERROR,
               "ACM_readACLservice",(*reasonCode_P)));
               status = LI_ERR_ACM;
            }
         }
         else /* if CSI_ASYNCHRON==li_MsgContext_P->ExchangeMode */
         {
            ret = ACM_readACLqueue(
                  li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName,
                  CSI_WRITE,&trgtACL,returnCode_P,reasonCode_P);
            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_ACM_ERROR,
               "ACM_readACLqueue",(*reasonCode_P)));
               status = LI_ERR_ACM;
            }
         }
      }

      /** Convert User Profiles string to a CDXRESULT structure **/

      if (LI_ERR_NONE == status)
      {
         ret = CDX_string2cdxresult(NULL,li_MsgContext_P->UserProfileList_P,
               &tmpUsPr, returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
            "CDX_string2cdxresult",(*reasonCode_P)));
            status = *reasonCode_P;
         }
      }

      /** Check ACLs **/

      if (LI_ERR_NONE == status)
      {
         ret = ACM_checkACL(&tmpUsPr,&trgtACL,returnCode_P);
         if (CSIFAILED==ret)
         {
            /*************************/
            /***** ACCESS DENIED *****/
            /*************************/

            EVR_SECUERRORLOG(EVR_USERERROR,(LI_MSG_ACCESS_DENIED,
               li_MsgContext_P->TraceInfo_P->OrgUserName, 
               (CSI_SYNCHRON == li_MsgContext_P->ExchangeMode ?
                li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName :
                li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName)) );

            status = LI_ERR_ACCESS_DENIED;
         }
         else
         {
            /* * * * * * * * * * * * **/
            /*     ACCESS GRANTED     */
            /** * * * * * * * * * * * */

            EVR_SECUTRACELOG(EVR_DEBUG, (LI_TRC_ACCESS_GRANTED,
               li_MsgContext_P->TraceInfo_P->OrgUserName,
               (CSI_SYNCHRON == li_MsgContext_P->ExchangeMode ?
                li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName :
                li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName)) );
         }
      }
      CDX_free(NULL,&trgtACL);
      CDX_free(NULL,&tmpUsPr);
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_ACMScheduler",*returnCode_P,*reasonCode_P);
   return(*returnCode_P);
}

/*****************************************************************************/
