/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
* 
* @FILE     li_assvc.c
* 
* @ABSTRACT
* Local Intelligence Asynchronous & Synchronous service routine.
*
* Warning: this code is to be compiled with the specific files:
* ------- 
* <@FILE li_synch.c> (Synchronous processing);
* <@FILE li_async.c> (Asynchronous processing);
*
* as well as the common files:
* <@FILE li_engin.c> (LI server common routines: init, term, context alloc);
* <@FILE li_opdis.c> (Operation Dispatcher);
* <@FILE li_msgro.c> (Message Routing Scheduler);
* <@FILE li_acmsc.c> (Access Control Management Scheduler);
* <@FILE li_prssc.c> (Data Presentation or Message Conversion Scheduler);
*
* This files contains the functions:
*
* @INCLUDE
*/

#include <stdio.h>
#include <stdlib.h>

/*************************************/
/*** GENERAL CCN/CSI INCLUDE FILES ***/
/*************************************/

#include "gcn_upcon.h"
#include "gcn_uptyp.h"

/** CSI API Include Files **/
#include "csi_upcon.h"
#include "csi_uptyp.h"
#include "csi_usext.h"

/** OS API Include Files **/
#include "osx_upcon.h"
#include "osx_upext.h"

/** CDX API Include Files **/
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"

/** GCN Tools Include Files **/
#include "gcn_upext.h"

/** PRES API Include Files **/
#include "prs_upcon.h"
#include "prs_upext.h"

/** EVENT REPORTING Include Files **/
#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "gcn_upmes.h"
#include "li_prmes.h"
#include "gcn_upmac.h"

/** QOS MACROS (Error handling) **/
#include "qos_upmac.h" 


/*************************************************/
/*** SPECIFIC LOCAL INTELLIGENCE INCLUDE FILES ***/
/*************************************************/

#include "li_prtyp.h"
#include "li_prcon.h"
#include "li_prext.h"
#include "li_upext.h"

/**************************************************/
/*** PROTOTYPES OF li_SyncProc AND li_AsyncProc ***/
/**************************************************/

extern CSILONG         li_SyncProc (
IN     LI_PROCCONTEXT *li_ProcContext_P,
INOUT  CSIPDU         **request_PP,
INOUT  LI_MSGCONTEXT  *li_MsgContext_P,
OUT    CSIPDU         **reply_PP,
OUT    CSILONG        *returnCode_P,
OUT    CSILONG        *reasonCode_P,
OUT    CSILONG        *svcRetCode_P
);

extern CSILONG         li_AsyncProc (
IN     LI_PROCCONTEXT *li_ProcContext_P,
INOUT  LI_MSGCONTEXT  *li_MsgContext_P,
OUT    CSILONG        *returnCode_P,
OUT    CSILONG        *reasonCode_P
);

/*******************************************************************/
/*** LI SERVICE ROUTINE FOR SYNCHRONOUS OR ASYNCHRONOUS MESSAGES ***/
/*** DO NOT USE FOR CONVERSATIONAL MESSAGES !!!                  ***/
/*******************************************************************/
/*
* @FUNCTION li_service
* 
* @DESCRIPTION The LI service (CSI service entry point)
*
* @ABSTRACT
* The li_Service is the "entry point" of the LI Engine. It is constructed 
* according to the CSI_service pattern, that is, its input parameter is a
* pointer to a <@TYPE CSISVCINFO>, and it ends with <@MACRODEF CSI_return>
* instead of the classical return.
* It is defined as extern, because it must be called by the underlying
* CSI layer.
* Warning: any Data Descriptor received by the LI engine is supposed to
*          contain some "Trace Information", either as a "normal" message,
*          or as a "trigger" message. This data is located in the Header
*          buffer of the CSIDD, before presentation-related info. For more
*          details on the structure of the Trace Info, please refer to the
*          GCN module functions <@FUNCTION gcn_GetTraceInfo> and 
*          <@FUNCTION gcn_BuildTraceInfo>. 
*
* This function:
* - allocates a message context with <@LOCALFUN li_AllocMsgContext>.
* - extracts the trace information with <@FUNCTION gcn_GetTraceInfo>.
* - according to the type of the message, the relevant processing is then
*   applied (either <@LOCALFUN li_SyncProc> or <@LOCALFUN li_AsyncProc>).
* - releases message context with <@LOCALFUN li_FreeMsgContext>.
*
* @RETURNVALUE CSIVOID
* The li_Service does not return any value, because it is similar to a
* Tuxedo service routine. The "main" is controlled by Tuxedo. 
* 
* @PARAMETER 
* IN     CSISVCINFO* svcInfo_P
* A pointer to the CSI service information structure.
*
* @SEEALSO
* <@TYPE CSISVCINFO>
* <@FUNCTION CSI_svrinit_of_the_LI> 
* <@FUNCTION CSI_svrdone_of_the_LI>
*/
CSIVOID li_service(CSISVCINFO *svcInfo_P)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   /** Declaration & Initialisation of the message context and  **/
   /** associated structures.                                   **/

   LI_MSGCONTEXT *li_MsgCntxt_P =  NULL;

   /** Declaration of the local variables for the function calls **/
   CSILONG     ret         = CSISUCCESS;       /* all purpose return code */
   CSILONG     returnCode  = CSISUCCESS;       /* idem, but as parameter  */
   CSILONG     reasonCode  = LI_ERR_NONE;      /* all purpose reason code */

   /** Declaration of pointers to the request parameters, namely: **/
   CSIPDU      *reqData_P   = NULL;  /* Pointer to the Request Data Descriptor */
   CSIPDU      *orgData_P   = NULL;  /* Original request data descriptor */
   CSIQOS     *reqQoS_P    = NULL;  /* Pointer to the Request QoS             */
 
   /** Declaration of the parameters of the final CSI_return function **/
   CSILONG     svcRetVal    = CSI_TPSUCCESS;    /* service Return Code */
   CSILONG     svcRetCode   = LI_ERR_NONE;      /* returned Error Code */
   CSIPDU      *replyData_P  = NULL;             /* returned Data Desc. */
   CSIQOS      replyQoS_S   = {CSIQOS_DEFAULT}; /* returned QoS        */
   CSILONG     retFlags     = 0L;               /* MANDATORY VALUE !!! */ 

   /** Miscellaneous parameters **/
   CSILONG     targetSvcRetCode = LI_ERR_NONE; /* Target service rcode */

   TRACE_FUNCTION_ENTRY("li_service");

   /*** PARAMETER VERIFICATION AND INITIALISATIONS ***/ 
   /**************************************************/

   /** Verification of the input CSISVCINFO pointer **/

   if (NULL==svcInfo_P)
   {
      EVR_ERRORLOG(EVR_CRITICAL,(LI_MSG_NULL_POINTER,"svcInfo_P"));
      svcRetVal  = CSI_TPFAIL;
      svcRetCode = LI_ERR_NULL_SVCINFO;
   }

   /** Verification of the fields of CSISVCINFO **/

   if (CSI_TPSUCCESS==svcRetVal)
   {
      if (strcmp(svcInfo_P->SvcName, li_ProcCntxt_S.SvcName)!=0)
      {
         EVR_ERRORLOG(EVR_CRITICAL,(LI_MSG_INVAL_PARAM,
         "svcInfo_P->SvcName") );
         svcRetVal = CSI_TPFAIL;
      }

      if (NULL == svcInfo_P->Request)
      {
         EVR_ERRORLOG(EVR_CRITICAL,(LI_MSG_NULL_POINTER,
         "svcInfo_P->Request") );
         svcRetVal = CSI_TPFAIL;
      }

      if (NULL == svcInfo_P->RequestQoS)
      {
         EVR_ERRORLOG(EVR_CRITICAL,(LI_MSG_NULL_POINTER,
         "svcInfo_P->RequestQoS") );
         svcRetVal = CSI_TPFAIL;
      }

      if (CSI_TPFAIL == svcRetVal) svcRetCode = LI_ERR_SVC_PARAM;
   }

   /** The information of the CSISVCINFO structure is copied to **/
   /** simplify the source code (easier to analyse and debug !) **/
   
   if (CSI_TPSUCCESS==svcRetVal)
   {
      reqData_P = (CSIPDU *)svcInfo_P->Request;    /* Copy of the DD  pointer       */
      orgData_P = reqData_P;
      reqQoS_P  = svcInfo_P->RequestQoS; /* Copy of the QOS pointer       */
   }

  /*
   * 1. Allocation of a LI message context.
   */
   /**************************************/

   if (CSI_TPSUCCESS==svcRetVal)
   {
      ret = li_AllocMsgContext(&li_MsgCntxt_P, &returnCode, &reasonCode);
      if (CSIFAILED==ret)
      {
         svcRetVal  = CSI_TPFAIL;
         svcRetCode = LI_ERR_ALLOC;  
      }
   }

   /*
   * 2. Extraction of the Trace Information from the Data Descriptor Header.
   */
   /***********************************************************************/

   if (CSI_TPSUCCESS==svcRetVal)
   { /* BEGIN EXTRACT INFO */
   ret = gcn_GetTraceInfo(GCN_GET,reqData_P,&(li_MsgCntxt_P->EventType),
         li_MsgCntxt_P->TraceInfo_P, li_MsgCntxt_P->TriggerQName,
         &(li_MsgCntxt_P->UserProfileList_P),
         &(li_MsgCntxt_P->PrevGatewayList_P),
         &returnCode, &reasonCode);
   if (CSIFAILED==ret)
   {
      svcRetVal  = CSI_TPFAIL;
      svcRetCode = LI_ERR_GCN_TRACE_INFO;
   }
   } /* ENDOF EXTRACT INFO */

   /*
   * 3. Copying the Quality of Service in the LI message context.
   */
   /************************************************************/

   if (CSI_TPSUCCESS==svcRetVal)
   { /* BEGIN: STORE QOS */
   memcpy(li_MsgCntxt_P->QoS_P,reqQoS_P,sizeof(CSIQOS) );
   li_MsgCntxt_P->SvcFlags = svcInfo_P->Flags;
   } /* ENDOF: STORE QOS */

   /*
   * 4. Selection of the suitable processing (SYNC/ASYNC).
   */
   /******************************************************/

   if (CSI_TPSUCCESS==svcRetVal)
   {/* BEGIN SWITCH */
   switch(li_MsgCntxt_P->EventType)
   {
      /** Case of "trigger message" from Trigger Monitor **/
      case CSI_TRIGGER_EVENT:
         li_MsgCntxt_P->ExchangeMode = CSI_ASYNCHRON;
         ret = li_AsyncProc(&li_ProcCntxt_S
                           ,li_MsgCntxt_P
                           ,&returnCode
                           ,&reasonCode
                           );
         break;

      /** Case of a "call" or "acall" from RAP/GAS  **/     
      case CSI_LOCAL_EVENT:
      /** Case of a "call" from a first LI          **/
      case CSI_REMOTE_EVENT:
      /** Case of a "call" or "acall" from http2csi **/     
      case CSI_DIRECT_REMOTE_EVENT:
         EVR_TRACELOG(EVR_DEBUG,(LI_TRC_SYNC_EVENT,
         li_MsgCntxt_P->TraceInfo_P->UniqueId,
         li_MsgCntxt_P->TraceInfo_P->OrgUserName,
         li_MsgCntxt_P->TraceInfo_P->OrgAppliName,
         li_MsgCntxt_P->TraceInfo_P->TrgtCCNServiceName) );
         
         li_MsgCntxt_P->ExchangeMode = CSI_SYNCHRON;

         ret = li_SyncProc(&li_ProcCntxt_S, &reqData_P, li_MsgCntxt_P,
                           &replyData_P,&returnCode,&reasonCode,
                           &targetSvcRetCode);

         if (CSIFAILED==ret)
	 {
            if (TUX_ERR_SVCFAIL==reasonCode)
            { /* The error has been signalled by the target server */
               reasonCode = targetSvcRetCode;
              /* Log the target service name */
              /* Service failure arror are no longer looged in error but 
               * only in AUDIT logs 
               */
              /*  EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_CALL_FAILED,"target",
                  li_MsgCntxt_P->TraceInfo_P->TrgtCCNServiceName,reasonCode)); */
            }
            else
            { /* The error occurred in the current stack */
               QOS_SET_CCN_ERR(li_MsgCntxt_P->QoS_P);
            }
         }
         break;

      default:
         /* Should not occur, as the event is checked in gcn_GetTraceInfo. */
         /* But one is never too cautious (belt and suspenders syndrome !) */
         ret        = CSIFAILED;
         reasonCode = LI_ERR_UNKNOWN_EVENT;
         break;
   }

   if (CSIFAILED == ret)
   {
      if (LI_ERR_ROLLBACK == reasonCode)
          svcRetVal = CSI_TPEXIT;
      else
          svcRetVal = CSI_TPFAIL;
      svcRetCode = reasonCode;
   }
   else
   {
      svcRetVal  = CSI_TPSUCCESS;
      svcRetCode = targetSvcRetCode;
   }

   } /* ENDOF SWITCH */

   /* 
   * 5. Copying the QoS from the LI message context into the replyQoS.
   */
   if (NULL != li_MsgCntxt_P)
   {
      memcpy(&replyQoS_S, li_MsgCntxt_P->QoS_P,sizeof(CSIQOS) );
   }

   /*
   * 6. Release of the LI message context.
   */
   ret = li_FreeMsgContext(&li_MsgCntxt_P, &returnCode,&reasonCode);
   if (CSIFAILED==ret)
   {
      svcRetVal  = CSI_TPFAIL;
      svcRetCode = reasonCode;
      QOS_SET_CCN_ERR((&replyQoS_S));
   }

   /* Release Data descriptor if needed */
   if (orgData_P != reqData_P)
   {
      /* Data Presentation has been applied, so reqData_P has to be
         released.  (orgData_P will be released in GTX_rturn function)
       */
      gcn_FreeDataDesc(&reqData_P, &returnCode, &reasonCode);
   }

   /*** CONCLUSION OF THE LI SERVICE: CSI_RETURN ***/
   /************************************************/

   /** Store Error Code in LI process context **/
   li_ProcCntxt_S.ErrNo = svcRetCode;

   TRACE_FUNCTION_EXIT("li_service",svcRetVal,svcRetCode);

   /** Return of the LI service **/
   CSI_return(svcRetVal,svcRetCode,(CSIDD *)replyData_P,&replyQoS_S,retFlags);
}

/*****************************************************************************/
