/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
* 
* @FILE     li_engin.c
* 
* @ABSTRACT
* Local Intelligence common ENGINe routines.
*
* This file contains the functions:
*
* @INCLUDE
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/*************************************/
/*** GENERAL CCN/CSI INCLUDE FILES ***/
/*************************************/

/** GCN Tools Include Files **/
#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upext.h"

/** CSI API Include Files **/
#include "csi_upcon.h"
#include "csi_uptyp.h"
#include "csi_usext.h"

/** OS API Include Files **/
#include "osx_upcon.h"
#include "osx_upext.h"

/** CDX API Include Files **/
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"

/** PRES API Include Files **/
#include "prs_upcon.h"
#include "prs_upext.h"

/** EVENT REPORTING Include Files **/
#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "gcn_upmes.h"
#include "li_prmes.h"
#include "gcn_upmac.h"

/*************************************************/
/*** SPECIFIC LOCAL INTELLIGENCE INCLUDE FILES ***/
/*************************************************/

#include "li_prtyp.h"
#include "li_prcon.h"
#include "li_prext.h"
#include "li_upext.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#ifdef _ASYNC
/*************************************************************************/
/* WARNING: In asynchronous mode, a LI server will enter a "get message" */
/* ======== loop during its CSI_svrinit(). This enables the async. LI to */
/*          empty its dedicated queue (defined via command-line CoT)     */
/*          without any call from the Trigger Monitor. The initial pur-  */
/*          pose of this was to prevent messages to linger in the queue  */
/*          when the LI crashes or is shut down during operations. When  */
/*          restarted, the async. LI resumes its work.                   */
/*          To activate this functionality, compile with the flag        */
/*          -D_ASYNC                                                     */
/*          DO NOT USE IN A CONVERSATIONAL LI (WON'T LINK, ANYWAY !)     */
/*          Of course, for this feature we need li_AsyncProc(), thus...  */
/*************************************************************************/

/********************************************/
/*** LI ASYNCHRONOUS PROCESSING PROTOTYPE ***/
/********************************************/
extern CSILONG         li_AsyncProc(
IN     LI_PROCCONTEXT *li_ProcContext_P,
INOUT  LI_MSGCONTEXT  *li_MsgContext_P,
OUT    CSILONG        *returnCode_P,
OUT    CSILONG        *reasonCode_P
);

#endif /* _ASYNC */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
* @GLOBALVAR LI_PROCCONTEXT li_ProcCntxt_S
* Declaration of the LI Process Context.
* It is used by different subcomponents of the LI engine, which are
* implemented in different source files.
* 
* @SEEALSO
* <@TYPE LI_PROCCONTEXT>
*/
/*********************************************************/
/*** LI PROCESS CONTEXT DECLARATION (with initializer) ***/
/*********************************************************/

LI_PROCCONTEXT li_ProcCntxt_S = {LI_PROCCONTEXT_DEFAULT};


/*******************************/
/*** ROUTINES IMPLEMENTATION ***/
/*******************************/

/*
* @FUNCTION CSI_svrinit_of_the_LI
*
* @DESCRIPTION Server initialization of the LI
*
* @ABSTRACT
* This function is used only at the start-up of the LI process (based on a
* Tuxedo server). It is called by Tuxedo (through the CSI stack). Its role
* is to initialise the server.
*
* Warning: the same code is used for the two types of executables, 
*          conversational and synchronous&asynchronous LIs. Thus, 
*          some parts are excluded with #ifdef. In particular, for a
*          sync-async LI, a specific portion of code handling the
*          processing of asynchronous messages in the LI queue is 
*          included (but not in the conv. LI).
*
* What it does:
* - Read the parameters (kind of a command-line options for the server) with
*   <@FUNCTION gcn_GetNextArg>;
* - Opens the logging system;
* - Initialises the OS with <@FUNCTION OS_init>;
* - Binds to the CDX with <@FUNCTION CDX_bind>;
* - Binds to the presentation module with <@FUNCTION PRES_bind>;
* - Binds to the CSI layer with <@FUNCTION CSI_bind>;
* - (sync-async only) Connects to the Queue Manager with 
*   <@FUNCTION CSI_mq_conn>;
* - Advertises the service with <@FUNCTION CSI_advertise>;
* - (asynchronous mode only) call <@LOCALFUN li_AsyncProc> to read the
*   messages lingering in the LI queue;
*
* @RETURNVALUE CSILONG
* The returnvalue is the same as the value pointed by returnCode_P (see below).
* It can be either CSISUCCESS or CSIFAILED.
*
* @PARAMETER 
* IN     CSICHAR256 svrArgs
*        This is the "argument string".
*
* OUT    CSILONG* returnCode_P
*        The value pointed by this pointer is the same as the 
*        returnvalue (see above).
*
* OUT    CSILONG* reasonCode_P
*        In case the return code is CSIFAILED, this reason code is an error
*        code (see below the possible values).
*
* @ERRORCODE CSI_ERR_ARG
* Invalid arguments were given (in the svrArgs string).
*
* @ERRORCODE CSI_ERR_POINTER
* Input parameters contain a NULL pointer.
*
* @SEEALSO
* <@FUNCTION CSI_svrdone_of_the_LI>
*/
/****************************************/
/*** LI SERVER INITIALISATION ROUTINE ***/
/****************************************/
CSILONG CSI_svrinit(
CSICHAR256  svrArgs, 
CSILONG    *returnCode_P,
CSILONG    *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATION ***/ 
   /***********************************/

   /** Pointers to global structures **/

   LI_PROCCONTEXT  *li_ProcCntxt_P = NULL; /* LI process context pointer */

   /** general working local variables **/

   CSILONG   ret        = CSISUCCESS;   /* All purpose return code !  */
   CSILONG   status     = LI_ERR_NONE;  /* Error status               */
       
   /** Local variables used to analyse the argument string **/

   CSIULONG    initPos = 0L;
   CSIULONG    argLen  = 0L;
   CSICHAR48   nextArg;

   /** Temporary local variables **/
   
   CSIQOS      tmpQoS_S   = {CSIQOS_DEFAULT};

   CSICHAR     *pcWkdDir;

   /** Specific variables (useful in asynchronous LIs) **/

   CSICHAR16      my_CoT      = {""};  /* Class of Traffic (in svrArgs) */
#ifdef _ASYNC
   CSILONG        my_Type     = 0L;    /* LI type (by CDX_getAppliType) */
   LI_MSGCONTEXT *li_MsgCtx_P = NULL;  /* A LI msg context (for ALIs)   */
   /* OR 4237 CSIMQHCONN  qMgrHandle = 0L; */
#endif

  /*** CHANGING DIRECTORY ***/
  /**************************/

   pcWkdDir = getenv("CCN_DAT_PATH");

   if(NULL != pcWkdDir)
   {
      chdir(pcWkdDir);
   }

  /*** INPUT POINTERS VERIFICATION ***/
  /***********************************/

   if ((NULL==returnCode_P)||(NULL==reasonCode_P))
   {
      return(CSIFAILED);
   }

   /*** INITIALISATIONS ***/
   /***********************/

   *returnCode_P = CSISUCCESS;         /* default values (normal case) */
   *reasonCode_P = LI_ERR_NONE;        /* idem                         */

    li_ProcCntxt_P = &li_ProcCntxt_S;   /* ...now pointing to the context */

   /*** EXTRACTION OF THE PARAMETERS OF THE ARGUMENT STRING ***/
   /***********************************************************/
   /** Rem: the current argument string is supposed to contain : **/
   /**      the <LI service name>                                **/
 
   /** First argument retrieval **/
 
   argLen = CSIMAXLEN_APPNAME;          /* max length of expected argument */
   ret = gcn_GetNextArg(svrArgs,GCN_SEPARATOR,&initPos,&argLen,nextArg,
                        returnCode_P,reasonCode_P);
   if ((CSIFAILED==ret)||(0==argLen))
   {
      *reasonCode_P = CSI_ERR_ARG;      /* MAJOR ERROR: the argument string */
      return(*returnCode_P);            /* has not the expected format.     */
   }

   strcpy(li_ProcCntxt_P->SvcName,nextArg); /* LI service */
   strcpy(li_ProcCntxt_P->Name   ,nextArg); /* Same as service name. */

   /** Second argument retrieval : Class Of Traffic expected **/

   initPos += argLen;
   argLen   = CSIMAXLEN_COTNAME;

   ret = gcn_GetNextArg(svrArgs,GCN_SEPARATOR,&initPos,&argLen, my_CoT,
                        returnCode_P, reasonCode_P);
   if ((CSIFAILED==ret)||(0==argLen))
   {
      *reasonCode_P = CSI_ERR_ARG;      /* MAJOR ERROR: the argument string */
      return(*returnCode_P);            /* has not the expected format.     */
   }

   /***  OPENING THE LOGGING SYSTEM ***/
   /***********************************/

   ret = EVR_OPENLOG(li_ProcCntxt_P->Name,"LI");
   if (CSIFAILED==ret)
   {
      *returnCode_P = CSIFAILED;
      *reasonCode_P = LI_ERR_NO_LOG;
      return(*returnCode_P);
   }

   /* Now that the logging system is supposed to be on, say hello ! */
   TRACE_FUNCTION_ENTRY("CSI_svrinit");

   /*** INITIALISATION OF THE OS CONTEXT ***/
   /****************************************/

   ret = OS_init(li_ProcCntxt_P->Name, returnCode_P, reasonCode_P);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_INIT_OS_ERROR, (*reasonCode_P)));
      status        = LI_ERR_OS_INIT;
   } 

   /*** CONNECTION TO THE CDX FACILITY ***/
   /**************************************/
   /** The LI server binds anonymously, hence the NULLs ... **/

   if (LI_ERR_NONE == status)
   { 
      ret = CDX_bind(NULL,NULL,NULL, returnCode_P, reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_INIT_CDX_ERROR, (*reasonCode_P)));
         status        = LI_ERR_CDX_BIND;
      }
   }

   /*** BINDING WITH THE PRESENTATION ***/
   /*************************************/

   if (LI_ERR_NONE == status)
   {
      ret = PRES_bind(li_ProcCntxt_P->Name, returnCode_P, reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_INIT_PRES_ERROR, (*reasonCode_P)));
         status        = LI_ERR_PRES_BIND;
      }
   }

   /*** BINDING WITH THE CSI LAYER ***/
   /**********************************/

   if (LI_ERR_NONE == status)
   { 
      ret = CSI_bind(li_ProcCntxt_P->Name,"",&tmpQoS_S, returnCode_P,
                     reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_INIT_CSI_ERROR, (*reasonCode_P)));
         status = *reasonCode_P;
      }
   }

/* in conversational mode, we do not need to connect to a queue manager,
 * since degraded mode is not allowed.
 */
#ifdef _ASYNC
   /*** CONNECTION TO THE LOCAL QUEUE MANAGER ***/
   /*********************************************/
   /** If no queue manager name is mentioned, the default local queue **/
   /** manager is selected.                                           **/

/* OR4237
   if (LI_ERR_NONE == status)
   {
      ret = CSI_mq_conn("",&qMgrHandle,returnCode_P,reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
         "CSI_mq_conn",(*reasonCode_P)));
         status = *reasonCode_P;
      }

      li_ProcCntxt_P->QMgrHandle = qMgrHandle; /+* Copy in the context *+/

   }
*/
#endif

   /*** ADVERTISING THE LI SERVICE ***/
   /**********************************/

   if (LI_ERR_NONE == status)
   {
      ret = CSI_advertise(li_ProcCntxt_P->SvcName,&li_service, returnCode_P,
                          reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
         "CSI_advertise",(*reasonCode_P)));
         status = *reasonCode_P;
      }
   }

   /*** RETRIEVING THE ENVIRONMENT VARIABLES: CCN_GATEWAY_NAME ***/
   /**************************************************************/

   if (LI_ERR_NONE == status)
   {
      ret = CDX_getLocalGateway(&(li_ProcCntxt_P->GatewayName),
            returnCode_P, reasonCode_P);
      if (CSIFAILED == ret)
      {
         EVR_ERRORLOG(EVR_MINOR,(LI_MSG_ENV_VAR_ERROR,
         "CCN_GATEWAY_NAME"));
         status        = LI_ERR_ENV_MISSING;
      }
   }


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#ifdef _ASYNC

   /****************************************/
   /* SPECIAL FEATURE: SEE WARNING ABOVE ! */
   /****************************************/

   /*** IDENTIFICATION OF THE LI, AND ACCORDINGLY, FIRST ASYNC.PROC. ***/
   /********************************************************************/

   /**************************************************************************/
   /* Programming note: in the following lines, the program goes to the      */
   /* next STEP only if previous step is OK.                                 */
   /* The code can be read as a check list, any failed action invalidating   */
   /* further processing.                                                    */
   /**************************************************************************/

   /* Retrieving the LI type, asynchronous or synchronous */

   if (LI_ERR_NONE == status)
   { /* BEGIN: STEP 0 OK */

   ret = CDX_getAppliType(NULL, li_ProcCntxt_P->Name, &my_Type, 
                          returnCode_P, reasonCode_P);
   if (CSISUCCESS == ret)
   { /* BEGIN: STEP 1 LI type identified OK */

   if ((CSI_ALI == my_Type) || (CSI_XLI == my_Type)) /* OR 3881 */
   { /* BEGIN: STEP 2 LI is ASYNCHRONOUS OK */

   /* Allocating a LI message context */

   ret = li_AllocMsgContext(&li_MsgCtx_P, returnCode_P, reasonCode_P);
   if (CSISUCCESS == ret)
   { /* BEGIN: STEP 3 LI message context allocated OK */

   /* Retrieves asynchronous LI queue name (with previously retrieved CoT) */

   ret = CDX_getCSINameOfCoT(NULL /* OR 3881: CSI_ALI */ 
                            ,my_Type
                            ,my_CoT
                            ,&li_MsgCtx_P->TriggerQName
                            ,returnCode_P
                            ,reasonCode_P
                            );
   if (CSISUCCESS == ret)
   { /* BEGIN: STEP 4 ALI queue name retrieved OK */

     li_MsgCtx_P->ExchangeMode = CSI_ASYNCHRON;

   /* Calling LI Asynchronous Processing for this queue */

   /* OR3917 BEGIN */

     /*
     li_AsyncProc(li_ProcCntxt_P, li_MsgCtx_P, returnCode_P, reasonCode_P);
     */

      /*** OR4237: CONNECTION TO THE LOCAL QUEUE MANAGER ***/
      /*****************************************************/
      /** As no queue manager name is mentioned, the default local queue **/
      /** manager is selected.                                           **/

      if (LI_ERR_NONE == status)
      {
         ret = CSI_mq_conn("",&(li_ProcCntxt_P->QMgrHandle),returnCode_P,reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
            "CSI_mq_conn",(*reasonCode_P)));
            status = *reasonCode_P;

           li_ProcCntxt_P->QMgrHandle = CSIMQHC_UNUSABLE_HCONN;
         }
      }

     if(CSISUCCESS != ALI_checkInputQ(li_ProcCntxt_P->QMgrHandle,
                                      li_MsgCtx_P->TriggerQName,
                                      li_ProcCntxt_P->SvcName,
                                      returnCode_P,
                                      reasonCode_P))
     {
        status = *reasonCode_P;
     }

     /*** OR4237: DISCONNECTING FROM THE LOCAL QUEUE MANAGER ***/
     /**********************************************************/

     if(CSIMQHC_UNUSABLE_HCONN != li_ProcCntxt_P->QMgrHandle)
     {
        ret = CSI_mq_disc(&(li_ProcCntxt_P->QMgrHandle),returnCode_P,reasonCode_P);
        if (CSIFAILED==ret)
        {
           EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
           "CSI_mq_disc",*reasonCode_P));
           status = *reasonCode_P;
        }
     }


   /* OR3917 END zzz */

   } /* ENDOF: STEP 4 */

   } /* ENDOF: STEP 3 */

   /* Releasing LI message context */

   if (NULL != li_MsgCtx_P)
   {
      li_FreeMsgContext(&li_MsgCtx_P, returnCode_P, reasonCode_P);
   }

   } /* ENDOF: STEP 2 */

   } /* ENDOF: STEP 1 */

   } /* ENDOF: STEP 0 */

#endif /* _ASYNC */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("CSI_svrinit",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}

 
/*****************************/
/*** LI SERVER TERMINATION ***/
/*****************************/
/*
* @FUNCTION CSI_svrdone_of_the_LI
*
* @DESCRIPTION Server termination of the LI
*
* @ABSTRACT
* This function is used only at the shutdown of the LI process (triggered by
* Tuxedo). It is called by Tuxedo (through the CSI stack). Its role is to ter-
* minate the different connections. 
*
* What it does:
* - (sync async only) disconnection from the queue manager with
*   <@FUNCTION CSI_mq_disc>;
* - unbinding from the CSI layer ( <@FUNCTION CSI_unbind> );
* - unbinding from the presentation module ( <@FUNCTION PRES_unbind> );
* - disconnection from CDX ( <@FUNCTION CDX_unbind> );
* 
* Note: no need to unadvertise the LI service as Tuxedo automatically
*       cleans up its table during the shutdown
*
* @RETURNVALUE CSIVOID
*
* @PARAMETER CSIVOID
*/
CSIVOID CSI_svrdone(CSIVOID)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   CSILONG ret        = CSISUCCESS;
   CSILONG status     = CSI_ERR_NONE;

   CSILONG returnCode = CSISUCCESS;
   CSILONG reasonCode = LI_ERR_NONE; 

   TRACE_FUNCTION_ENTRY("CSI_svrdone");

   /*****************************************************************/
   /* No need to unadvertise the LI service as Tuxedo automatically */
   /* cleans up its table during the shutdown.                      */
   /*****************************************************************/

#ifdef _ASYNC
   /*** DISCONNECTING FROM THE LOCAL QUEUE MANAGER ***/
   /**************************************************/
/* OR4237
   ret = CSI_mq_disc(&(li_ProcCntxt_S.QMgrHandle),&returnCode,&reasonCode);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
      "CSI_mq_disc",reasonCode));
      status = reasonCode;
   } 
*/
#endif

   /*** UNBINDING FROM THE CSI LAYER ***/
   /************************************/

   ret = CSI_unbind(&returnCode,&reasonCode);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CSI_ERROR,
      "CSI_unbind",reasonCode));
      status = reasonCode;
   }

   /*** UNBINDING FROM THE PRES MODULE ***/
   /**************************************/

   ret = PRES_unbind(&returnCode,&reasonCode);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_PRES_ERROR,
      "PRES_unbind",reasonCode));
      status = reasonCode;
   }

   /*** DISCONNECTING FROM THE CDX FACILITY ***/
   /*******************************************/

   ret = CDX_unbind(NULL, &returnCode, &reasonCode);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
      "CDX_unbind",reasonCode));
      status = reasonCode;
   }

   /*** THE END !!! ***/
   /*******************/

   if (CSI_ERR_NONE != status) returnCode = CSIFAILED;
                          else returnCode = CSISUCCESS;
   reasonCode = status;

   /*** SAY GOODBYE ! ***/
   /*********************/

   TRACE_FUNCTION_EXIT("CSI_svrdone",returnCode,reasonCode);

   /*** TERMINATING THE OS CONTEXT ***/
   /**********************************/

   OS_term(&returnCode,&reasonCode);

   /***************************************************************/
   /*** NOTE: WHATEVER THE RESULTS OF THE PREVIOUS INSTRUCTIONS ***/
   /***       AFTER THE RETURN THE TUXEDO MAIN WILL STOP THE    ***/
   /***       SERVER.                                           ***/
   /***************************************************************/
   return;
}


/***************************************/
/*** ALLOCATION OF A MESSAGE CONTEXT ***/
/***************************************/
/*
* @LOCALFUN li_AllocMsgContext
*
* @DESCRIPTION Allocates a message context
*
* @ABSTRACT
* This function allocates a new LI message context, <@TYPE LI_MSGCONTEXT>.
* Actually, a message context is allocated once for each service call to the
* LI. Even in the case of the asynchronous LI, which processes several 
* messages in one single service execution, only one message context is 
* allocated (and then reused). 
* 
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* INOUT  LI_MSGCONTEXT **li_MsgContext_PP
*        The currently processed message context.
*
* OUT    CSILONG *returnCode_P
*        The returnCode.
*
* OUT    CSILONG *reasonCode_P
*        The reasonCode.
*
* @SEEALSO
* <@LOCALFUN li_FreeMsgContext>
* <@TYPE LI_MSGCONTEXT>
*/
CSILONG li_AllocMsgContext(
INOUT   LI_MSGCONTEXT   **li_MsgContext_PP,
OUT     CSILONG          *returnCode_P,
OUT     CSILONG          *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   /** Miscellaneous local variables **/

   CSILONG       ret             = CSISUCCESS;
   CSILONG       status          = LI_ERR_NONE;

   /** Message context associated variables **/

   LI_MSGCONTEXT li_MsgContext_S = {LI_MSGCONTEXT_DEFAULT};

   CSIIITFHDR    traceInfo_S     = {CSIIITFHDR_DEFAULT};
   CSIIITFHDR   *traceInfo_P     = NULL;

   CSIQOS        msgQoS_S        = {CSIQOS_DEFAULT};
   CSIQOS       *msgQoS_P        = NULL;

   CSIMQMD       msgDesc_S       = {CSIMQMD_DEFAULT};
   CSIMQMD      *msgDesc_P       = NULL;

   TRACE_FUNCTION_ENTRY("li_AllocMsgContext");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_AllocMsgContext",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** OS ALLOCATION FOR THE STRUCTURES (QOS,MQMD,TRACEINFO) ***/
   /*************************************************************/

   if (LI_ERR_NONE == status)
   { 
      ret = OS_mem_alloc(sizeof(CSIIITFHDR),(CSIVOID**)&traceInfo_P,
            returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
         "OS_mem_alloc(traceInfo_P)",(*reasonCode_P)));
         status        = LI_ERR_ALLOC;
      }
   }

   if (LI_ERR_NONE == status)
   { 
      ret = OS_mem_alloc(sizeof(CSIQOS),(CSIVOID**)&msgQoS_P,
            returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         CSILONG lLocalReturn, lLocalReason;
         (CSIVOID) OS_mem_free((CSIVOID**)&traceInfo_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
         "OS_mem_alloc(msgQoS_P)",(*reasonCode_P)));
         status        = LI_ERR_ALLOC;
      }
   }

   if (LI_ERR_NONE == status)
   { 
      ret = OS_mem_alloc(sizeof(CSIMQMD),(CSIVOID**)&msgDesc_P,
            returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         CSILONG lLocalReturn, lLocalReason;
         (CSIVOID) OS_mem_free((CSIVOID**)&traceInfo_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         (CSIVOID) OS_mem_free((CSIVOID**)&msgQoS_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
         "OS_mem_alloc(msgDesc_P)",(*reasonCode_P)));
         status        = LI_ERR_ALLOC;
      }
   }

   if (LI_ERR_NONE == status)
   { 
      ret = OS_mem_alloc(sizeof(LI_MSGCONTEXT),(CSIVOID**)li_MsgContext_PP,
            returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         CSILONG lLocalReturn, lLocalReason;
         (CSIVOID) OS_mem_free((CSIVOID**)&traceInfo_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         (CSIVOID) OS_mem_free((CSIVOID**)&msgQoS_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         (CSIVOID) OS_mem_free((CSIVOID**)&msgDesc_P
                              ,&lLocalReturn
                              ,&lLocalReason
                              );
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
         "OS_mem_alloc(li_MsgContext_P)",(*reasonCode_P)));
         status        = LI_ERR_ALLOC;
      }
   }

   /*** ASSIGNATIONS (INITIALISATIONS) ***/
   /**************************************/

   if (LI_ERR_NONE == status)
   { 
      memcpy(*li_MsgContext_PP,&li_MsgContext_S, sizeof(li_MsgContext_S));
      memcpy(traceInfo_P,&traceInfo_S,sizeof(traceInfo_S));
      memcpy(msgQoS_P,&msgQoS_S, sizeof(msgQoS_S));
      memcpy(msgDesc_P,&msgDesc_S, sizeof(msgDesc_S));
      (*li_MsgContext_PP)->TraceInfo_P = traceInfo_P;
      (*li_MsgContext_PP)->QoS_P       = msgQoS_P;
      (*li_MsgContext_PP)->MsgDesc_P   = msgDesc_P;
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_AllocMsgContext",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}


/************************************/
/*** RELEASE OF A MESSAGE CONTEXT ***/
/************************************/
/*
* @LOCALFUN li_FreeMsgContext
*
* @DESCRIPTION Releases a message context
*
* @ABSTRACT
* This functions releases a LI message context. Symmetrical to 
* <@LOCALFUN li_AllocMsgContext>.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* INOUT  LI_MSGCONTEXT **li_MsgContext_PP
*        The currently processed message context.
*
* OUT    CSILONG *returnCode_P
*        The returnCode.
*
* OUT    CSILONG *reasonCode_P
*        The reasonCode.
*
* @SEEALSO
* <@LOCALFUN li_AllocMsgContext>
* <@TYPE LI_MSGCONTEXT>
*/
CSILONG li_FreeMsgContext(
INOUT   LI_MSGCONTEXT    **li_MsgContext_PP,
OUT     CSILONG           *returnCode_P,
OUT     CSILONG           *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATION ***/
   /***********************************/

   CSILONG ret    = CSISUCCESS;
   CSILONG status = LI_ERR_NONE; 

   TRACE_FUNCTION_ENTRY("li_FreeMsgContext");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_FreeMsgContext",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** RELEASE ALL STRUCTURES ***/
   /******************************/
   ret = OS_mem_free((CSIVOID**)&((*li_MsgContext_PP)->TraceInfo_P),
         returnCode_P, reasonCode_P);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
      "OS_mem_free(TraceInfo_P)",(*reasonCode_P)));
      status        = LI_ERR_FREE;
   }

   ret = OS_mem_free((CSIVOID**)&((*li_MsgContext_PP)->QoS_P),
         returnCode_P, reasonCode_P);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
      "OS_mem_free(QoS_P)",(*reasonCode_P)));
      status        = LI_ERR_FREE;
   }

   ret = OS_mem_free((CSIVOID**)&((*li_MsgContext_PP)->MsgDesc_P),
         returnCode_P, reasonCode_P);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
      "OS_mem_free(MsgDesc_P)",(*reasonCode_P)));
      status        = LI_ERR_FREE;
   }

   ret = gcn_ReleaseInfo(&((*li_MsgContext_PP)->UserProfileList_P),
                         &((*li_MsgContext_PP)->PrevGatewayList_P),
                         returnCode_P, reasonCode_P );
   if (CSIFAILED == ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
      "gcn_ReleaseInfo",(*reasonCode_P)));
      status = LI_ERR_FREE;
   }

   ret = OS_mem_free((CSIVOID**)li_MsgContext_PP,returnCode_P,reasonCode_P);
   if (CSIFAILED==ret)
   {
      EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
      "OS_mem_free(li_MsgContext_P)",(*reasonCode_P)));
      status        = LI_ERR_FREE;
   }

   /*** THE END ***/
   /***************/
   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_FreeMsgContext",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}

/****************************************************************************/
