/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
*
* @FILE     li_msgro.c
*
* @ABSTRACT
* The Local Intelligence MeSsaGe ROuting scheduler.
* This file contains the functions:
*
* @INCLUDE
*/


#include <stdio.h>
#include <string.h>

#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upext.h"

#include "osx_upcon.h"
#include "osx_upext.h"

#include "rou_upcon.h"
#include "rou_upext.h"

#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"
#include "cdx_upmac.h"

#include "csi_upcon.h"

#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "gcn_upmes.h"
#include "li_prmes.h"

#include "li_prcon.h"
#include "li_prtyp.h"
#include "li_prext.h"
#include "li_upext.h"
#include "gcn_upmac.h"

/************************************/
/*** LI MESSAGE ROUTING SCHEDULER ***/
/************************************/
/*
* @LOCALFUN li_MsgRoutScheduler
*
* @DESCRIPTION Performs message routing
*
* @ABSTRACT
* The Message Routing Scheduler routine. This function is called by 
* <@LOCALFUN li_OperationDispatcher> to perform the routing task. It relies
* on the Message Routing (ROU) module. At this point, we only know whether
* the message comes from an local application or a distant gateway (it was
* passed along in the trace information, and is kept in the EventType field of
* <@TYPE LI_MSGCONTEXT>).
* 
* The executed actions are:
* - (local msg only) the name of the current gateway is added to the
*   gateway lists;
* - retrieving the name of the target gateway with
*   <@FUNCTION rou_GetTargetGatewayName>;
* - determinate whether this gateway is the local gateway or not
* - [current restriction] reject message if more than 2 gateways are
*   involved (N gateways routing still looming, but...!).
* - (remote target only) add target gateway name to the gateway list;
* - getting the "next hop" name (service to be called, queue
*   to be used...) with <@FUNCTION CDX_getCSINameOfCoT>.
* - Updating the value of the EventType according to the results of the
*   routing.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* IN     LI_PROCCONTEXT *li_ProcContext_P
*        The Local Intelligence process context.
*
* INOUT  LI_MSGCONTEXT  *li_MsgContext_P
*        The currently processed message context.
*
* OUT    CSIPDU         *dataDesc_P
*        The message data descriptor
*
* OUT    CSILONG        *returnCode_P
*        The returnCode.
*
* OUT    CSILONG *reasonCode_P
*        The reasonCode.
*
* @SEEALSO
* <@LOCALFUN li_OperationDispatcher>
*/
CSILONG li_MsgRoutScheduler (
IN      LI_PROCCONTEXT   *li_ProcContext_P,
INOUT   LI_MSGCONTEXT    *li_MsgContext_P,
OUT     CSIPDU           *dataDesc_P,
OUT     CSILONG          *returnCode_P,
OUT     CSILONG          *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/
 
   CSILONG    ret           = CSISUCCESS;
   CSILONG    status        = LI_ERR_NONE;
   CSILONG    tmpReas       = LI_ERR_NONE;

   CSILONG    toLocal       = TRUE;

   CSILONG    typeLI        = 0L;

   CSIULONG   argLength     = 0L;
   CSIULONG   initPos       = 0L;


   CSIULONG   tmpFlag = 0x00000000;

   TRACE_FUNCTION_ENTRY("li_MsgRoutScheduler");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_MsgRoutScheduler",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /*** ADDING THE CURRENT GATEWAY TO THE LIST ***/
   /**********************************************/
   /** Only if the message is local (i.e. CSI_LOCAL_EVENT) **/

   if (LI_ERR_NONE == status)
   {
      if ( (CSI_LOCAL_EVENT==li_MsgContext_P->EventType) &&
           (li_MsgContext_P->TraceInfo_P->NbOfPrevGateways == 0L) )
      { /* ===> Message sent by a LOCAL application */
         if (NULL==li_MsgContext_P->PrevGatewayList_P)
         {
            ret = OS_mem_alloc(GCN_MAX_PREVGW_LEN,
                  (CSIVOID**)&(li_MsgContext_P->PrevGatewayList_P),
                  returnCode_P,reasonCode_P);
            if (CSIFAILED == ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_OS_ERROR,
               "OS_mem_alloc (PrevGatewayList_P)",(*reasonCode_P)));
               status = LI_ERR_ALLOC;
            }
            else
            {
               li_MsgContext_P->PrevGatewayList_P[0] = GCN_END_STRING_CHAR;
               li_MsgContext_P->OrigGateway[0] = GCN_END_STRING_CHAR;
            }
         }

         if (LI_ERR_NONE == status)
         {
            ret = gcn_AddArgToBuf(li_ProcContext_P->GatewayName,
                  GCN_MAX_PREVGW_LEN,GCN_SEPARATOR,
                  &(li_MsgContext_P->PrevGatewayList_P),NULL,NULL,
                  returnCode_P,reasonCode_P);
            if (CSIFAILED==ret)
            {
               EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
               "gcn_AddArgToBuf",(*reasonCode_P)));
               status = LI_ERR_INTERNAL;
            }

            li_MsgContext_P->TraceInfo_P->NbOfPrevGateways++;
         }
      }
      else
      {  /* ===> Message sent by a REMOTE application */
         /* Getting originator gateway name (for audit) */

         initPos = 0L;
         argLength = GCN_MAX_PREVGW_LEN;
         ret = gcn_GetNextBufArg(li_MsgContext_P->PrevGatewayList_P,
                                 GCN_SEPARATOR, &initPos,
                                 &argLength, li_MsgContext_P->OrigGateway,
                                 returnCode_P, reasonCode_P);
         if (CSIFAILED == ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_GetNextArgBufArg",(*reasonCode_P)));
            status = LI_ERR_INTERNAL;
         }
      }
   }

   /*** RETRIEVING THE TARGET GATEWAY NAME ***/
   /******************************************/


   if (LI_ERR_NONE == status)
   {
      CSICHAR48 ressourceName;

      memset(ressourceName, '\0', sizeof ressourceName);
      if (CSI_SYNCHRON==li_MsgContext_P->ExchangeMode)
      {
         GCN_STRCPY(ressourceName, li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName);
      }
      else
      {
         GCN_STRCPY(ressourceName, li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName);
      }

      if('\0' != ressourceName[0])
      {
         ret = rou_GetTargetGatewayName(ressourceName
                                       ,li_MsgContext_P->RunMode
                                       ,li_MsgContext_P->ExchangeMode
                                       ,li_MsgContext_P->ExchangeMode == CSI_SYNCHRON
                                            ? CSITRUE
                                            : (li_MsgContext_P->MsgDesc_P->MsgType == CSIMQMT_REPORT
                                                   ? CSIFALSE
                                                   : CSITRUE
                                              )
                                       ,&li_MsgContext_P->DestGateway
                                       ,returnCode_P
                                       ,&tmpReas
                                       );

         if (CSIFAILED==ret)
         {
              CSICHAR *tmpStr;
              switch(tmpReas)
              {
              case(ROU_ERR_INVALID_SITE_NAME):
                 *reasonCode_P = LI_ERR_INVALID_SITE_NAME;
                 tmpStr = "Invalid SITE name";
                 break;

              case(ROU_ERR_INVALID_SVC_NAME):
                 *reasonCode_P = LI_ERR_INVALID_SVC_NAME;
                 tmpStr = "Invalid SERVICE name";
                 break;

              case(ROU_ERR_INVALID_QUE_NAME):
                 *reasonCode_P = LI_ERR_INVALID_QUE_NAME;
                 tmpStr = "Invalid QUEUE name";
                 break;

              case(ROU_ERR_INVALID_APP_NAME):
                 *reasonCode_P = LI_ERR_INVALID_APP_NAME;
                 tmpStr = "Invalid APPLICATION name";
                 break;

              case(ROU_ERR_INVALID_MODE):
                 *reasonCode_P = LI_ERR_INVALID_MODE;
                 tmpStr = "Invalid RUN MODE";
                 break;

              case(ROU_ERR_UNAVAILABLE_QUE):
                 *reasonCode_P = LI_ERR_UNAVAILABLE_QUE;
                 tmpStr = "Unavailable QUEUE";
                 break;

              case(ROU_ERR_UNAVAILABLE_SVC):
                 *reasonCode_P = LI_ERR_UNAVAILABLE_SVC;
                 tmpStr = "Unavailable SERVICE";
                 break;
   
              case(ROU_ERR_UNAVAILABLE_APP):
                 *reasonCode_P = LI_ERR_UNAVAILABLE_APP;
                 tmpStr = "Unavailable APPLICATION";
                 break;

              case(ROU_ERR_UNAVAILABLE_GTW):
                 *reasonCode_P = LI_ERR_UNAVAILABLE_GTW;
                 tmpStr = "Unavailable GATEWAY";
                 break;

              case(ROU_ERR_UNDETERMINATE):
                 *reasonCode_P = LI_ERR_ROU_UNDETERMINATE;
                 tmpStr = "UNDETERMINATE";
                 break;

              default:
                 *reasonCode_P = tmpReas;
                 tmpStr = "item";
                 break;
           }

           EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_RES_ERROR,
           tmpStr,ressourceName));

           status = *reasonCode_P;

        }
      }
      else
      {
         *returnCode_P = CSIFAILED;
         *reasonCode_P = LI_ERR_EMPTY_RESOURCE_NAME;

         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_EMPTY_RESOURCE_NAME));
         status = *reasonCode_P;
      }
   }


   /* Identify the case: the Target Gateway is either  */
   /*                    the local one or a remote one */


   if (LI_ERR_NONE == status)
   {
      ret = CDX_isLocalGateway(li_MsgContext_P->DestGateway,&toLocal,
                               returnCode_P,reasonCode_P);
      if (CSIFAILED == ret)
      {
         status = *reasonCode_P;
      }
   }

   /*** FOR NOW, THE MESSAGE CAN AT MOST TRANSIT THROUGH 2 ***/
   /*** GATEWAYS. LET'S CHECK IT.                          ***/
   /**********************************************************/

   if (LI_ERR_NONE == status)
   {
      if (  (FALSE==toLocal)
          &&(  (CSI_REMOTE_EVENT       ==li_MsgContext_P->EventType)
             ||(CSI_DIRECT_REMOTE_EVENT==li_MsgContext_P->EventType)
            )
         )
      {
         status = LI_ERR_LOOP_DETECTED;
      }
   }

   /*** MESSAGE LOOPING DETECTION ***/
   /*********************************/
   /** Only if the message is remote (i.e. CSI_REMOTE_EVENT) **/
   /** and current GW is not destination.                    **/
   /** Not used currently (previous section causes shortcut) **/

   if (LI_ERR_NONE == status)
   {
      if (   (   (CSI_REMOTE_EVENT       ==li_MsgContext_P->EventType)
              || (CSI_DIRECT_REMOTE_EVENT==li_MsgContext_P->EventType)
             )
          && (FALSE==toLocal)
         )
      {
         ret = li_LoopDetection(li_MsgContext_P->DestGateway,
               li_MsgContext_P->TraceInfo_P->NbOfPrevGateways,
               li_MsgContext_P->PrevGatewayList_P, returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            status = *reasonCode_P;
         }
      }
   }

   /*** ADDING THE TARGET GATEWAY TO THE LIST ***/
   /*********************************************/
   /** Only if destination is not local **/

   if (LI_ERR_NONE == status)
   {
      if (FALSE==toLocal)
      {
         ret = gcn_AddArgToBuf(li_MsgContext_P->DestGateway,
               GCN_MAX_PREVGW_LEN,GCN_SEPARATOR,
               &(li_MsgContext_P->PrevGatewayList_P),NULL,NULL,
               returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_AddArgToBuf",(*reasonCode_P)));
            status = LI_ERR_INTERNAL;
         }

         li_MsgContext_P->TraceInfo_P->NbOfPrevGateways++;
      }
   }

   if (LI_ERR_NONE == status)
   {
   /* Initialize the XferMemory (li_MsgContext) */
   /*-------------------------------------------*/
      li_MsgContext_P->XferMemory = 0; 

      /*       set flags XferMemory in MsgContextcould         */
      /*           indicating to the LI that the transfer      */
      /*           should be by memory and not by file (if needed)  */

      if  (  (CSI_REMOTE_EVENT       ==li_MsgContext_P->EventType)
           ||(CSI_DIRECT_REMOTE_EVENT==li_MsgContext_P->EventType)
          )
      {
          li_MsgContext_P->XferMemory += CSI_DD_FLAGS_XFER_BY_MEMORY;
      }
   }

   /*** RETRIEVING THE "NEXT HOP" NAME ***/
   /**************************************/


   if (LI_ERR_NONE == status)
   {
      if (TRUE==toLocal)
      {
         if (CSI_SYNCHRON==li_MsgContext_P->ExchangeMode)
         {
            strncpy(li_MsgContext_P->NextHopName,
                    li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                    sizeof(CSICHAR48));
         }
         else
         {
            strncpy(li_MsgContext_P->NextHopName,
                    li_MsgContext_P->TraceInfo_P->TrgtCCNQueueName,
                    sizeof(CSICHAR48));
         }
      }
      else
      { 
         if (CSI_SYNCHRON==li_MsgContext_P->ExchangeMode)
         {
               typeLI = CSI_RLI;
         }

         if (CSI_ASYNCHRON==li_MsgContext_P->ExchangeMode) 
         {
               typeLI = CSI_XLI;
         }

         /*  Test the Version and                                 */
         /*       select LI as Next hop name                      */
         /*       set flags XferMemory in MsgContextcould         */
         /*           indicating to the LI that the transfer      */
         /*           should be by memory and not by file (if needed)  */

         ret = CDX_getCSINameOfCoT(li_MsgContext_P->DestGateway
                                  ,typeLI
                                  ,li_MsgContext_P->QoS_P->CoT
                                  ,&li_MsgContext_P->NextHopName
                                  ,returnCode_P
                                  ,reasonCode_P
                                  );

         li_MsgContext_P->XferMemory += CSI_DD_FLAGS_XFER_BY_MEMORY;

         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
            "CDX_getCSINameOfCoT",(*reasonCode_P)));
            status = LI_ERR_CDX_GET_COT_NAME;
         }
      }
   }

   /*** UPDATING THE VALUE OF EVENT TYPE ACCORDING TO THE DESTINATION ***/
   /*********************************************************************/

   if (LI_ERR_NONE == status)
   {
      if (FALSE == toLocal) 
      {
         li_MsgContext_P->EventType = CSI_REMOTE_EVENT;
      }
      else
      {
         li_MsgContext_P->EventType = CSI_LOCAL_EVENT;
      }
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_MsgRoutScheduler",*returnCode_P,*reasonCode_P);
   return(*returnCode_P);
}


/*********************************/
/*** MESSAGE LOOPING DETECTION ***/
/*********************************/
/*
* @LOCALFUN li_LoopDetection
*
* @DESCRIPTION Checks the gateway list to detect any loop in the routing.
*
* @ABSTRACT
* This function is part of the Message Routing Scheduler of the LI. It
* checks any possible loop in the routing of the message; that is, if the
* current gateway name is spotted amongst the lists of the previous gateways,
* a LI_ERR_LOOP_DETECTED error code is returned.
*
* [Note: since a routing currently involves at most two gateways, this
*        function is not in use... for now.                            ]
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* IN     CSICHAR256 localGateway
*        The Distinguished Name of the Local Gateway (as defined in the
*        environment variable CCN_GATEWAY_NAME for instance).
*
* IN     CSIULONG nbOfPrevGateways
*        The number of gateway names in the following prevGatewayBuf_P buffer.
*
* IN     CSICHAR *prevGatewayBuf_P
*        An argument buffer (string of arguments separated by GCN_SEPARATOR)
*        containing the names of the gateways through the message transited.
*
* OUT    CSILONG *returnCode_P
*        The returnCode.
*
* OUT    CSILONG *reasonCode_P
*        The reasonCode.
*
* @SEEALSO
* <@LOCALFUN li_MsgRoutScheduler>
*/
CSILONG li_LoopDetection (
IN      CSICHAR256      localGateway,
IN      CSIULONG        nbOfPrevGateways,
IN      CSICHAR        *prevGatewayBuf_P,
OUT     CSILONG        *returnCode_P,
OUT     CSILONG        *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   CSILONG     ret        = CSISUCCESS;
   CSILONG     status     = LI_ERR_NONE;

   CSIULONG    initPos    = 0L;
   CSICHAR256  argWord    = {""};
   CSIULONG    argLength  = 0L;
   CSIULONG    idx        = 0L;

   TRACE_FUNCTION_ENTRY("li_LoopDetection");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_LoopDetection",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   if (0 == nbOfPrevGateways)
   {
      TRACE_FUNCTION_EXIT("li_LoopDetection",*returnCode_P,*reasonCode_P);
      return(*returnCode_P);
   }

   /*** COMPARE ALL GATEWAY NAMES TO CURRENT GATEWAY NAME ***/
   /*********************************************************/

   if (LI_ERR_NONE == status)
   {
      initPos = 0L;

      for (idx = 0; idx < nbOfPrevGateways; ++idx)
      {
         argLength = GCN_MAX_PREVGW_LEN;
         ret = gcn_GetNextBufArg(prevGatewayBuf_P,GCN_SEPARATOR,&initPos,
               &argLength,argWord,returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_GetNextBufArg",(*reasonCode_P)));
            status = LI_ERR_INTERNAL;
         }

         if (strcmp(argWord,localGateway)==0)
         {
            EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_LOOP_DETECTED));
            status = LI_ERR_LOOP_DETECTED;
         }

         initPos += argLength;
      }
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
                         else *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   TRACE_FUNCTION_EXIT("li_LoopDetection",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}

/*****************************************************************************/
